/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <assert.h>
#include "stash.h"
#include "main.h"
#include "mem.h"
#include "inst.h"
#include "predict.h"
#include "smt.h"

#define FP_OFFSET 32
#define FPCR 64

extern context *thecontexts[];
extern int debug;
int debugsync;

int simulating;
extern int numcontexts;
extern unsigned long cyc;

int activethreads = 1;
int multiplethreads = 0;

/* Per processor lock used to implement LDx_L and STx_C */
int lock = 0;  
unsigned long locked_address = 0;

/*  All instruction emulation is done via these short routines
    that assume that a bunch of static information about the
    instruction is known and is stored in the stash data
    structure.
*/

void
emulate_unimplemented_instruction(context *current, StashData *stash)
{
	printf("ERROR: unimplemented instruction\n");
	printf("ERROR: pc = %lx, opcode = %x\n",
		current->pc, (stash->inst >> 26));
	printf("ERROR: instruction = %x, cycle = %ld\n",
		stash->inst, cyc);
}

void
emulate_invalid_instruction(context *current, StashData *stash)
{
	printf("ERROR: invalid (illegal) instruction\n");
	printf("ERROR: pc = %lx, opcode = %x\n",
		current->pc, (stash->inst >> 26));
	printf("ERROR: instruction = %x, cycle = %ld\n",
		stash->inst, cyc);
}

void
emulate_nop(context *current, StashData *stash)
{
	return;
}

void
emulate_lda(context *current, StashData *stash)
{
	{
		*stash->RA = *stash->RB + SEXT(stash->offset);
		current->R[31] = 0;
	}
}

void
emulate_ldah(context *current, StashData *stash)
{
	{
		*stash->RA = *stash->RB + SEXT32(65536*stash->offset);
		current->R[31] = 0;
	}
}

void
emulate_ldq_u(context *current, StashData *stash)
{
	unsigned long va = (*stash->RB + SEXT(stash->offset)) & ~7;

	{
	  if (current->wrong_path || current->follow_sync)
	    *stash->RA = safe_read_mem_quad(current, va);
	  else
	    *stash->RA = read_mem_quad(current, va);
	  current->R[31] = 0;
	}

	current->srcmem = va;
}

void
emulate_stq_u(context *current, StashData *stash)
{
	unsigned long va = (*stash->RB + SEXT(stash->offset)) & ~7;

	if (!(current->wrong_path || current->follow_sync))
	{
		set_mem_quad(current, va, *stash->RA);
	}

	current->destmem = va;
}

void
emulate_inta_addl(context *current, StashData *stash)
{
	{
		*stash->RC = SEXT32( (*stash->RA + *stash->RB) & 0xffffffff);
		current->R[31] = 0;
	}
}

void
emulate_inta_addq(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA + *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_inta_cmple(context *current, StashData *stash)
{
	{
		*stash->RC =
			((signed long)*stash->RA <= (signed long)*stash->RB);
		current->R[31] = 0;
	}
}

void
emulate_inta_cmpult(context *current, StashData *stash)
{
	{
		*stash->RC =
			((unsigned long)*stash->RA < (unsigned long)*stash->RB);
		current->R[31] = 0;
	}
}

void
emulate_inta_s4addl(context *current, StashData *stash)
{
	{
		*stash->RC =
			SEXT32((((*stash->RA) <<2) + *stash->RB) & 0xffffffff);
		current->R[31] = 0;
	}
}

void
emulate_inta_s4subq(context *current, StashData *stash)
{
	{
		*stash->RC = ((*stash->RA) <<2) - *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_inta_s8subl(context *current, StashData *stash)
{
	{
		*stash->RC =
			SEXT32((((*stash->RA) <<3) - *stash->RB) & 0xffffffff);
		current->R[31] = 0;
	}
}

void
emulate_inta_cmpbge(context *current, StashData *stash)
{
	unsigned long rav;
	unsigned long rbv;
	int i;

	{
		rav = *stash->RA;
		rbv = *stash->RB;

		*stash->RC = 0;

		for (i = 56; i >= 0; i -= 8)
		{
			*stash->RC = *stash->RC << 1;
			*stash->RC |= (
				(rav >> i & (long)0xff) >=
				(rbv >> i & (long)0xff));
		}
		current->R[31] = 0;
	}
}

void
emulate_inta_cmplt(context *current, StashData *stash)
{
	{
		*stash->RC =
			((signed long) *stash->RA < (signed long) *stash->RB);
		current->R[31] = 0;
	}
}

void
emulate_inta_subl(context *current, StashData *stash)
{
	{
		*stash->RC =
			SEXT32( (*stash->RA - *stash->RB) & 0xffffffff);
		current->R[31] = 0;
	}
}

void
emulate_inta_subq(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA - *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_inta_s4addq(context *current, StashData *stash)
{
	{
		*stash->RC = ((*stash->RA) <<2) + *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_inta_s8addl(context *current, StashData *stash)
{
	{
		*stash->RC =
			SEXT32(((*stash->RA<<3) + *stash->RB) & 0xffffffff);
		current->R[31] = 0;
	}
}

void
emulate_inta_s8subq(context *current, StashData *stash)
{
	{
		*stash->RC = ((*stash->RA) <<3) - *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_inta_cmpeq(context *current, StashData *stash)
{
	{
		*stash->RC = (*stash->RA == *stash->RB);
		current->R[31] = 0;
	}
}

void
emulate_inta_cmpule(context *current, StashData *stash)
{
	{
		*stash->RC =
			((unsigned long) *stash->RA <=
			(unsigned long) *stash->RB);
		current->R[31] = 0;
	}
}

void
emulate_inta_s4subl(context *current, StashData *stash)
{
	{
		*stash->RC =
			SEXT32((((*stash->RA) <<2) - *stash->RB) & 0xffffffff);
		current->R[31] = 0;
	}
}

void
emulate_inta_s8addq(context *current, StashData *stash)
{
	{
		*stash->RC = ((*stash->RA) <<3) + *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_intl_and(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA & *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_intl_bic(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA & ~*stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_intl_bis_smt_fork(context *current, StashData *stash)
{
	int i;
	int j;

	/*
	** Find a free context and get it going.
	*/
	if (current->wrong_path || current->follow_sync)
	  return;
	simulating = 1;

#ifdef SMT_DEBUG2
	if (debugsync)
	{
		printf("Thread running status:\t{");

		for (i = 0; i < numcontexts; i++)
		{
			if (thecontexts[i]->running)
			{
	      			printf("%dR ",i);
	    		}
			else
			{
				printf("%dN ",i);
			}
		}

		printf("}\n");
	}
#endif

	{
	  for (i = 0; i < numcontexts; i++)
	    {
	      if (!thecontexts[i]->running 
		  && thecontexts[i]->masterid == current->masterid)
		{
		  thecontexts[i]->running = 1;
		  break;
		}
	    }

		if (i == numcontexts)
		{
			printf("ERROR: Couldn't find free ");
			printf("HW context for SMT_FORK\n");
		}
#ifdef SMT_DEBUG2
		else if (debugsync)
		{
			printf("\tSMT_FORK for current context (#%d)\n",
			current->id);

	      		printf("using free HW context #%d\n", i);
		}
#endif
		/*
		** Fork begins execution at the function label passed
		** as a parameter
		*/
		for(j = 0; j < 32; j++)
		{
			thecontexts[i]->R[j] = current->R[j];
			thecontexts[i]->FPR[j] = current->FPR[j];
		}

#ifdef MFLOW
		thecontexts[i]->nextpc = current->R[4];
#else
		thecontexts[i]->nextpc = current->R[REG_A2];
#endif
		/*
		** thecontexts[i]->R[30] = current->stacktop;
		*/
		thecontexts[i]->R[30] = thecontexts[i]->stacktop;

		activethreads++;
		multiplethreads = 1;


#ifdef SMT_DEBUG2
#ifdef MFLOW
		if (debugsync)
		{
			printf("Setting pc for new context #%d to 0x%lx\n",
				i, current->R[4]);
		}
#else
		if (debugsync)
		{
			printf("Setting pc for new context #%d to 0x%lx\n",
				i, current->R[REG_A2]);
		}
#endif
#endif
	}
}

void
emulate_intl_bis_smt_getid(context *current, StashData *stash)
{
	{
#ifdef SMT_DEBUG2
		if (debugsync)
		{
			printf("\tSMT_GETID for current context (#%d)\n",
				current->id);
		}
#endif

		/*
		** Place the ID in the function return register.
		*/
#ifdef MFLOW
		current->R[2] = current->id;
#else
		current->R[REG_V0] = current->id;
#endif
	}
}

void
emulate_intl_bis_smt_createstack(context *current, StashData *stash)
{
	{
#ifdef SMT_DEBUG2
		if (debugsync)
		{
			printf("\tSMT_CREATESTACK for current context (#%d)\n", 
				current->id);
		}
#endif
		current->R[30] = current->stacktop;
	}
}

void
emulate_intl_bis_smt_hw_lock(context *current, StashData *stash)
{
	/*
	** This instruction attempts to acquire the lock at
	** the memory location 0($16) or 0($2).  If the lock
	** is already set, then it blocks and the acquire 
	** data structure is updated with this information.
	*/

#ifdef MFLOW
	current->srcmem = current->destmem = current->R[2];
#else
	current->srcmem = current->destmem = current->R[REG_A0];
#endif
}

void
emulate_intl_bis_smt_release(context *current, StashData *stash)
{
	/*
	** This instruction gives up the lock at 
	** the memory location 0($16) or 0($2).
	*/

#ifdef MFLOW
	current->destmem = current->R[2];
#else
	current->destmem = current->R[REG_A0];
#endif
}

void
emulate_intl_bis_smt_terminate(context *current, StashData *stash)
{
	/*
	** Stop executing this thread
	*/
  if (!(current->wrong_path || current->follow_sync)) 
	{	
		current->running = 0;

		if (--activethreads == 1)
		{
		  /*counting = 0;*/
		}
	}
}

void emulate_read_if_full(context *current, StashData *stash)
{
}

void emulate_write_and_fill(context *current, StashData *stash)
{
}

void
emulate_intl_bis_smt_start_sim(context *current, StashData *stash)
{
	simulating = 1;

	zero_cstats();
	zero_pstats();
}

void
emulate_intl_bis_smt_print_sim(context *current, StashData *stash)
{
  if (!current->wrong_path && !current->follow_sync)
    {
      printf("********  SMT_PRINT_SIM Timing Point **********\n");
      schedstats();
      print_cstats();
      predict_stats();
      reset_stats();
    }
}

void
emulate_intl_bis_smt_end_sim(context *current, StashData *stash)
{
	{
		simulating = 0;
	}
}

/* obsolete */
void
emulate_intl_bis_smt_tas(context *current, StashData *stash)
{
	/*
	** This instruction does the following sequence atomically
	**
	**	ldl	$0, 0($16)
	**	bis	$31, 1, $reg
	**	stl	$reg, 0($16)
	**  
	** For the multiflow calling convention, it is:
	**
	**	ldl	$temp, 0($2)
	**	bis	$31, 1, $reg
	** 	stl	$reg, 0($2)
	**	bis	$temp, $temp, $2
	*/

#ifdef MFLOW
	current->srcmem = current->destmem = current->R[2];
#else
	current->srcmem = current->destmem = current->R[REG_A0];
#endif
}

/* FIXME -- deleted smt_restart */

void
emulate_intl_bis(context *current, StashData *stash)
{
  *stash->RC = *stash->RA | *stash->RB;
  current->R[31] = 0;
}

void
emulate_intl_cmoveq(context *current, StashData *stash)
{
	{
		if (*stash->RA == 0)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_cmovlbc(context *current, StashData *stash)
{
	{
		if ((*stash->RA & 1) == 0)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_cmovlbs(context *current, StashData *stash)
{
	{
		if (*stash->RA & 1)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_cmovge(context *current, StashData *stash)
{
	{
		if ((signed long) *stash->RA >= 0)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}


void
emulate_intl_cmovgt(context *current, StashData *stash)
{
	{
		if ((signed long) *stash->RA > 0)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_cmovle(context *current, StashData *stash)
{
	{
		if ((signed long) *stash->RA <= 0)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_cmovlt(context *current, StashData *stash)
{
	{
		if ((signed long) *stash->RA < 0)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_cmovne(context *current, StashData *stash)
{
	{
		if (*stash->RA)
		{
			*stash->RC = *stash->RB;
			current->R[31] = 0;
		}
	}
}

void
emulate_intl_eqv(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA ^ ~*stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_intl_ornot(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA | ~*stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_intl_xor(context *current, StashData *stash)
{
	{
		*stash->RC = *stash->RA ^ *stash->RB;
		current->R[31] = 0;
	}
}

void
emulate_ints_extbl(context *current, StashData *stash)
{
	{
		unsigned long temp = (*stash->RA >> ((*stash->RB & 0x7) * 8));
		*stash->RC = temp & (long) 0xff;
		current->R[31] = 0;
	}
}

void
emulate_ints_extlh(context *current, StashData *stash)
{
	{
		unsigned long temp =
			(*stash->RA << (64 - (*stash->RB & 0x7) * 8));
		*stash->RC =
			temp & (long) 0xffffffff;
		current->R[31] = 0;
	}
}

void
emulate_ints_extll(context *current, StashData *stash)
{
	{
		unsigned long temp = (*stash->RA >> ((*stash->RB & 0x7) * 8));
		*stash->RC = temp & (long) 0xffffffff;
		current->R[31] = 0;
	}
}

void
emulate_ints_extqh(context *current, StashData *stash)
{
	{
		*stash->RC = (*stash->RA << (64 - (*stash->RB & 0x7) * 8));
		current->R[31] = 0;
	}
}

void
emulate_ints_extql(context *current, StashData *stash)
{
	{
		*stash->RC = (*stash->RA >> ((*stash->RB & 0x7) * 8));
		current->R[31] = 0;
	}
}

void
emulate_ints_extwh(context *current, StashData *stash)
{
	{
		unsigned long temp =
			(*stash->RA << (64 - (*stash->RB & 0x7) * 8));
		*stash->RC =
			temp & (long) 0xffff;
		current->R[31] = 0;
	}
}

void
emulate_ints_extwl(context *current, StashData *stash)
{
	    unsigned long temp = (*stash->RA >> ((*stash->RB & 0x7) * 8));
	    *stash->RC = temp & (long) 0xffff;
	    current->R[31] = 0;
}

void
emulate_ints_insbl(context *current, StashData *stash)
{
	    unsigned long temp = (*stash->RA << ((*stash->RB & 0x7) * 8));
	    *stash->RC = temp & ((unsigned long) 0xff << ((*stash->RB & 0x7) * 8));
	    current->R[31] = 0;
}

void
emulate_ints_inslh(context *current, StashData *stash)
{
	    *stash->RC = (*stash->RA >> (63 - (*stash->RB & 0x7) * 8))>>1 &
	      ((unsigned long)0xffffffff >> ((8 - (*stash->RB & 0x7)) * 8));
	    current->R[31] = 0;
}

void
emulate_ints_insll(context *current, StashData *stash)
{
	    unsigned long temp = (*stash->RA << ((*stash->RB & 0x7) * 8));
	    *stash->RC = temp & ((unsigned long)0xffffffff << ((*stash->RB & 0x7) * 8));
	    current->R[31] = 0;
}

void
emulate_ints_insqh(context *current, StashData *stash)
{
    *stash->RC = (*stash->RA >> (63 - (*stash->RB & 0x7) * 8))>>1 &
      ((unsigned long)0xffffffffffffffff >> ((8 - (*stash->RB & 0x7)) * 8));
    current->R[31] = 0;
}

void
emulate_ints_insql(context *current, StashData *stash)
{
    *stash->RC = (*stash->RA << ((*stash->RB & 0x7) * 8)) &
      ((unsigned long)0xffffffffffffffff << ((*stash->RB & 0x7) * 8));
    current->R[31] = 0;
}

void
emulate_ints_inswh(context *current, StashData *stash)
{
    *stash->RC = (*stash->RA >> (63 - (*stash->RB & 0x7) * 8))>>1 &
      ((unsigned long)0xffff >> ((8 - (*stash->RB & 0x7)) * 8));
    current->R[31] = 0;
}

void
emulate_ints_inswl(context *current, StashData *stash)
{
      unsigned long temp = (*stash->RA << ((*stash->RB & 0x7) * 8));

	*stash->RC = temp & ((unsigned long)0xffff << ((*stash->RB & 0x7) * 8));
	current->R[31] = 0;
}

void
emulate_ints_mskbl(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xff << ((*stash->RB & 0x7) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_msklh(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xffffffff >> ((8 - (*stash->RB & 0x7)) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_mskll(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xffffffff << ((*stash->RB & 0x7) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_mskqh(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xffffffffffffffff >> ((8 - (*stash->RB & 0x7)) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_mskql(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xffffffffffffffff << ((*stash->RB & 0x7) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_mskwh(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xffff >> ((8 - (*stash->RB & 0x7)) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_mskwl(context *current, StashData *stash)
{
	    unsigned long temp = ((unsigned long) 0xffff << ((*stash->RB & 0x7) * 8));
	    *stash->RC = *stash->RA & ~temp;
	    current->R[31] = 0;
}

void
emulate_ints_sll(context *current, StashData *stash)
{
    *stash->RC = (*stash->RA) << (*stash->RB & 0x3f);
    current->R[31] = 0;
}

void
emulate_ints_sra(context *current, StashData *stash)
{
    *stash->RC =  (signed long) *stash->RA >> (*stash->RB & 0x3f);
    current->R[31] = 0;
}

void
emulate_ints_srl(context *current, StashData *stash)
{
    *stash->RC = (*stash->RA) >> (*stash->RB & 0x3f);
    current->R[31] = 0;
}

void
emulate_ints_zap(context *current, StashData *stash)
{

	    unsigned long temp = (long) 0xff;
	    unsigned long rav = *stash->RA;
	    unsigned long rbv = *stash->RB;
            int i;
	    *stash->RC = 0;
	    for (i=1; i<0x100; i=i<<1) {
	      if (i & ~rbv)
		*stash->RC |= rav & temp;
	      temp = temp << 8;
	    }
	    current->R[31] = 0;
}

void
emulate_ints_zapnot(context *current, StashData *stash)
{
	    unsigned long temp = (long) 0xff;
	    unsigned long rav = *stash->RA;
	    unsigned long rbv = *stash->RB;
            int i;
	    *stash->RC = 0;
	    for (i=1; i<0x100; i=i<<1) {
	      if (i & rbv)
		*stash->RC |= rav & temp;
	      temp = temp << 8;
	    }
	    current->R[31] = 0;
}

void
emulate_intm_mull(context *current, StashData *stash)
{
	    *stash->RC = SEXT32( (*stash->RA * *stash->RB) & 0xffffffff);
	    current->R[31] = 0;
}

void
emulate_intm_mulq(context *current, StashData *stash)
{
	    *stash->RC = *stash->RA * *stash->RB;
	    current->R[31] = 0;
}

#define ARITH_OVFL(RESULT, OP1, OP2) (RESULT < OP1 || RESULT < OP2)

long
high_multiply (long v1, long v2)
{
  register unsigned long a, b, c, d;
  register unsigned long bd, ad, cb, ac;
  register unsigned long mid, mid2, carry_mid = 0;

  a = (v1 >> 32) & (long) 0xffffffff;
  b = v1 & (long) 0xffffffff;
  c = (v2 >> 32) & (long) 0xffffffff;
  d = v2 & (long) 0xffffffff;

  bd = b * d;
  ad = a * d;
  cb = c * b;
  ac = a * c;

  mid = ad + cb;
  if (ARITH_OVFL (mid, ad, cb))
    carry_mid = 1;

  mid2 = mid + ((bd >> 32) & (long) 0xffffffff);
  if (ARITH_OVFL (mid2, mid, ((bd >> 32) & (long) 0xffffffff)))
    carry_mid += 1;

return ac + (carry_mid << 32) + ((mid2 >> 32) & (long) 0xffffffff);
}

void
emulate_intm_umulh(context *current, StashData *stash)
{
	    *stash->RC = high_multiply(*stash->RA, *stash->RB);
	    current->R[31] = 0;
}

void
emulate_fltl_cpys(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  current->FWR[rc] = ( current->FWR[ra] & (unsigned long)1<<63) |
	    (current->FWR[rb] & 0x7fffffffffffffff);
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_cpyse(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	current->FWR[rc] = ( current->FWR[ra] & (unsigned long) 0xfff<<52) |
	  ( current->FWR[rb] & (unsigned long) 0xfffffffffffff);
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_cpysn(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	current->FWR[rc] = (current->FWR[ra]>>63 ^ 1) << 63 |
	  (current->FWR[rb] & 0x7fffffffffffffff);
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_cvtlq(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	long longhold = current->FWR[rb];
	int inthold = ((longhold>>32) & 0xc0000000) | ((longhold>>29) & 0x3fffffff);
	(signed long) current->FWR[rc] = (signed) inthold;
}

void
emulate_fltl_cvtql(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	long longhold = current->FWR[rb];
	current->FWR[rc] = ((longhold>>32) & 0xc000000000000000) | ((longhold) & 0x3fffffff)<<29;
}

void
emulate_fltl_fcmoveq(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	if ((current->FWR[ra] << 1) == 0) {
	  current->FPR[rc] = current->FPR[rb];
	  current->FPR[31] = 0;
	}
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_fcmovge(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	if (((current->FWR[ra] << 1) == 0) || (current->FWR[ra]>>63 == 0)) {
	  current->FPR[rc] = current->FPR[rb];
	  current->FPR[31] = 0;
	}
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_fcmovgt(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	if (((current->FWR[ra] << 1) != 0) && (current->FWR[ra]>>63 == 0)) {
	  current->FPR[rc] = current->FPR[rb];
	  current->FPR[31] = 0;
	}
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_fcmovle(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	if (((current->FWR[ra] << 1) == 0) || (current->FWR[ra]>>63)) {
	  current->FPR[rc] = current->FPR[rb];
	  current->FPR[31] = 0;
	}
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_fcmovlt(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	if (((current->FWR[ra] << 1) != 0) && (current->FWR[ra]>>63)) {
	  current->FPR[rc] = current->FPR[rb];
	  current->FPR[31] = 0;
	}
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_fcmovne(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	if ((current->FWR[ra] << 1) != 0) {
	  current->FPR[rc] = current->FPR[rb];
	  current->FPR[31] = 0;
	}
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_fltl_mt_fpcr(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	current->fpcr = current->FPR[ra];
	stash->desthaz = FPCR;
	stash->srchaz1 = ra + FP_OFFSET;
}

void
emulate_fltl_mf_fpcr(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  current->FPR[ra] = current->fpcr;
	  current->FPR[31] = 0;
	  stash->srchaz1 = FPCR;
	  stash->desthaz = ra + FP_OFFSET;
}

void
emulate_flti_add(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  current->FPR[rc] = current->FPR[ra] + current->FPR[rb];
	  current->FPR[31] = 0;
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_flti_cmpteq(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;
  unsigned int rb = (stash->inst >> 16) & 0x1f;
  unsigned int rc = stash->inst & 0x1f;
  
    if ((current->FWR[ra] == current->FWR[rb]) ||
	(current->FWR[ra]<<1 == 0 && current->FWR[rb]<<1 == 0))
      current->FPR[rc] = 2;
    else current->FPR[rc] = 0;
    current->FPR[31] = 0;
    stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_flti_cmptlt(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;
  unsigned int rb = (stash->inst >> 16) & 0x1f;
  unsigned int rc = stash->inst & 0x1f;
  
    if (current->FPR[ra] < current->FPR[rb])
      current->FPR[rc] = 2;
    else current->FPR[rc] = 0;
    current->FPR[31] = 0;
    stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_flti_cmptle(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;
  unsigned int rb = (stash->inst >> 16) & 0x1f;
  unsigned int rc = stash->inst & 0x1f;
  
    if (current->FPR[ra] <= current->FPR[rb])
      current->FPR[rc] = 2;
    else current->FPR[rc] = 0;
    current->FPR[31] = 0;
    stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_flti_cvtq(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  (double) current->FPR[rc] = (signed long)current->FWR[rb];
	  current->FPR[31] = 0;
}

void
emulate_flti_cvtts(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  current->FPR[rc] = (float) current->FPR[rb];
	  current->FPR[31] = 0;
}

void
emulate_flti_divs(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;
  unsigned int rb = (stash->inst >> 16) & 0x1f;
  unsigned int rc = stash->inst & 0x1f;
  
  if (current->wrong_path && current->FPR[rb] == 0)
    current->FPR[rc] = 0.0;
  else {
    current->FPR[rc] = current->FPR[ra] / current->FPR[rb];
    current->FPR[31] = 0;
    stash->srchaz2 = ra + FP_OFFSET;
    stash->delay = 17;
  }
}

void
emulate_flti_divt(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  if (current->wrong_path && current->FPR[rb] == 0)
	    current->FPR[rc] = 0.0;
	  else {
	    current->FPR[rc] = current->FPR[ra] / current->FPR[rb];
	    current->FPR[31] = 0;
	    stash->srchaz2 = ra + FP_OFFSET;
	    stash->delay = 30;
	  }
}

void
emulate_flti_mul(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  current->FPR[rc] = current->FPR[ra] * current->FPR[rb];
	  current->FPR[31] = 0;
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_flti_sub(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	  current->FPR[rc] = current->FPR[ra] - current->FPR[rb];
	  current->FPR[31] = 0;
	stash->srchaz2 = ra + FP_OFFSET;
}

void
emulate_flti_cvttq(context *current, StashData *stash)
{
	unsigned int ra = (stash->inst >> 21) & 0x1f;
	unsigned int rb = (stash->inst >> 16) & 0x1f;
	unsigned int rc = stash->inst & 0x1f;

	long longhold;

	(signed long) longhold = current->FPR[rb];
	(signed long) current->FWR[rc] = (signed long) longhold;
}

void
emulate_jmp(context *current, StashData *stash)
{
	mem_addr wrong;

	  current->nextpc = *stash->RB & ~3;

      stash->taken = 1;
      if (current->wrong_path || current->follow_sync) {
	current->nextpc = get_btblookup(current->pc, current->masterid);
	if (current->nextpc == 0) {
	  current->nextpc = current->pc + 4;
	  stash->taken = 0;
	}
      }
      else {
	current->prediction = btblookup(current->pc, current->nextpc,
					&wrong, current->masterid,1,JUMP);
	if (current->prediction == PREDICT_WRONG && simulating) {
	  current->rightpc = current->nextpc;
	  current->nextpc = wrong;
	  if (wrong == 0) {
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	}
	*stash->RA = current->pc + 4;
	current->R[31] = 0;
      }
}

void
emulate_jsr(context *current, StashData *stash)
{
	mem_addr wrong;

	stash->taken = 1;
	current->nextpc = *stash->RB & ~3;
      if (current->wrong_path || current->follow_sync) {
	rs_push(current, current->pc + 4);
	current->nextpc = get_btblookup(current->pc, current->masterid);
	if (current->nextpc == 0) {
	  current->nextpc = current->pc + 4;
	  stash->taken = 0;
	}
      }
      else {
	current->prediction =  btblookup(current->pc, current->nextpc,
					 &wrong, current->masterid,1,JUMP);
	rs_push(current, current->pc + 4);
	if (current->prediction == PREDICT_WRONG && simulating) {
	  current->rightpc = current->nextpc;
	  current->nextpc = wrong;
	  if (wrong == 0) {
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	}
	*stash->RA = current->pc + 4;
	current->R[31] = 0;
      }
}

void
emulate_retn(context *current, StashData *stash)
{
	mem_addr wrong;

	stash->taken = 1;
	current->nextpc = *stash->RB & ~3;
      if (current->wrong_path || current->follow_sync) {
	current->nextpc = wp_rs_pop(current);
	if (current->nextpc == 0) {
	  current->nextpc = current->pc + 4;
	  stash->taken = 0;
	}
      }
      else {
	current->misfetch = 1 - btblookup(current->pc, current->nextpc,
					  &wrong, current->masterid,1,RETURN);
	current->prediction = rs_pop(current, current->nextpc, &wrong);
	if (current->prediction == PREDICT_WRONG && simulating) {
	  current->rightpc = current->nextpc;
	  current->nextpc = wrong;
	  if (wrong == 0) {
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	}
	else if (current->misfetch && simulating) {
	  current->rightpc = current->nextpc;
	  current->nextpc = current->pc + 4;
	  stash->taken = 0;
	}
	*stash->RA = current->pc + 4;
	current->R[31] = 0;
      }
}

void
emulate_jsr_coroutine(context *current, StashData *stash)
{
	mem_addr wrong;

	stash->taken = 1;
	current->nextpc = *stash->RB & ~3;
      if (current->wrong_path || current->follow_sync) {
	current->nextpc = wp_rs_pop(current);
	if (current->nextpc == 0) {
	  current->nextpc = current->pc + 4;
	  stash->taken = 0;
	}
	rs_push(current, current->pc + 4);
      }
      else {
	current->misfetch = 1 - btblookup(current->pc, current->nextpc,
					  &wrong, current->masterid,1,RETURN);
	current->prediction = rs_pop(current, current->nextpc, &wrong);
	rs_push(current, current->pc + 4);
	if (current->prediction == PREDICT_WRONG && simulating) {
	  current->rightpc = current->nextpc;
	  current->nextpc = wrong;
	  if (wrong == 0) {
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	}
	else if (current->misfetch && simulating) {
	  current->rightpc = current->nextpc;
	  current->nextpc = current->pc + 4;
	  stash->taken = 0;
	}
	*stash->RA = current->pc + 4;
	current->R[31] = 0;
      }
}

void
emulate_lds(context *current, StashData *stash)
{
  unsigned int    ra = (stash->inst >> 21) & 0x1f;
  unsigned int    rc = stash->inst & 0x1f;
  unsigned long va = (*stash->RB + SEXT(stash->offset));
  long longhold, e1, e2;

  if (current->wrong_path || current->follow_sync)
    longhold = safe_read_mem_word(current, va);
  else
    longhold = read_mem_word(current, va);
  e1 = longhold & 0x40000000;
  e2 = (longhold>>23) & 0x7f;
  if (e1) {
    if (e2 == 0x3f800000)
      e2 = 0x7ff;
    else
      e2 |= 0x400;
  }
  else {
    if (e2 == 0)
      e2 = 0;
    else
      e2 |= 0x380;
  }
  current->FWR[ra] = (longhold & 0x80000000)<<32 |
    e2<<52 | (longhold & 0x7fffff)<<29;
  current->srcmem =  va;
}

void
emulate_ldt(context *current, StashData *stash)
{
  unsigned long va = (*stash->RB + SEXT(stash->offset));
  unsigned int    ra = (stash->inst >> 21) & 0x1f;
  if (current->wrong_path || current->follow_sync)
    current->FWR[ra] = safe_read_mem_quad(current, va);
  else
    current->FWR[ra] =  read_mem_quad(current, va);
  current->srcmem = va;
}

void
emulate_sts(context *current, StashData *stash)
{
	unsigned long va = (*stash->RB + SEXT(stash->offset));
 unsigned int    ra = (stash->inst >> 21) & 0x1f;

      long longhold = current->FWR[ra];
      int inthold = ((longhold>>32) & 0xc0000000) | ((longhold>>29) & 0x3fffffff);
      if (!(current->wrong_path || current->follow_sync))
	set_mem_word(current, va, inthold);
    current->destmem = va;
}

void
emulate_stt(context *current, StashData *stash)
{
  unsigned long va = (*stash->RB + SEXT(stash->offset));
  unsigned int    ra = (stash->inst >> 21) & 0x1f;
  
  if (!(current->wrong_path || current->follow_sync))
    set_mem_quad(current, va, current->FWR[ra]);
  current->destmem = va;
}

emulate_ldl(context *current, StashData *stash)
{
	unsigned long va = (*stash->RB + SEXT(stash->offset));

	if (current->wrong_path || current->follow_sync)
	  (signed long) *stash->RA = 
	    (signed int) safe_read_mem_word(current, va);
	else
	  (signed long) *stash->RA = (signed int) read_mem_word(current, va);
	current->R[31] = 0;
	current->srcmem = va;
}

emulate_ldq(context *current, StashData *stash)
{
	unsigned long va = (*stash->RB + SEXT(stash->offset));

	if (current->wrong_path || current->follow_sync)
	  *stash->RA = safe_read_mem_quad(current, va);
	else
	  *stash->RA = read_mem_quad(current, va);
	current->R[31] = 0;
	current->srcmem = va;
}

/*ldl_l and ldq_l */
emulate_ldlq_l(context *current, StashData *stash)
{
	current->srcmem = (*stash->RB + SEXT(stash->offset));
}


emulate_stl(context *current, StashData *stash)
{
    unsigned long va = (*stash->RB + SEXT(stash->offset));

    if (!(current->wrong_path || current->follow_sync))
      set_mem_word(current, va, *stash->RA & (unsigned long) 0xffffffff);
    current->destmem = va;
}

emulate_stq(context *current, StashData *stash)
{
	unsigned long va = (*stash->RB + SEXT(stash->offset));

    if (!(current->wrong_path || current->follow_sync))
      set_mem_quad(current, va,*stash->RA);
    current->destmem = va;
}

/* stl_c and stq_c */
emulate_stlq_c(context *current, StashData *stash)
{
	current->destmem = (*stash->RB + SEXT(stash->offset));
}

emulate_bsr(context *current, StashData *stash)
{
        stash->taken = 1;
	if (current->wrong_path || current->follow_sync) {
	  rs_push(current, current->pc + 4);
	  current->nextpc = get_btblookup(current->pc, current->masterid);
	  if (current->nextpc == 0) {
	    stash->taken = 0;
	    current->nextpc = current->pc + 4;
	  }
	}
	else {
	  current->nextpc = current->pc+(SEXT21(stash->inst & 0x1fffff) << 2)+4;
	  current->misfetch = 1 - btblookup(current->pc, current->nextpc, 0,
					    current->masterid, 1, BRANCH);
	  if (current->misfetch && simulating) {
	    current->rightpc = current->nextpc;
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	  rs_push(current, current->pc + 4);
	  *stash->RA = current->pc + 4;
	  current->R[31] = 0;
	}
}

emulate_br(context *current, StashData *stash)
{
        stash->taken = 1;
	if (current->wrong_path || current->follow_sync) {
	  current->nextpc = get_btblookup(current->pc, current->masterid);
	  if (current->nextpc == 0) {
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	}
	else {
	  current->nextpc = current->pc+(SEXT21(stash->inst & 0x1fffff) << 2)+4;
	  *stash->RA = current->pc + 4;
	  current->R[31] = 0;
	  current->misfetch = 1 - btblookup(current->pc, current->nextpc,
					    0, current->masterid, 1, BRANCH);
	  if (current->misfetch && simulating) {
	    current->rightpc = current->nextpc;
	    current->nextpc = current->pc + 4;
	    stash->taken = 0;
	  }
	}
}

emulate_fbeq(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;

  if (current->FPR[ra] == 0.0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_fblt(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;

  if (current->FPR[ra] < 0.0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_fble(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;

  if (current->FPR[ra] <= 0.0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_fbne(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;

  if (current->FPR[ra] != 0.0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_fbge(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;

  if (current->FPR[ra] >= 0.0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_fbgt(context *current, StashData *stash)
{
  unsigned int ra = (stash->inst >> 21) & 0x1f;

  if (current->FPR[ra] > 0.0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_blbc(context *current, StashData *stash)
{
  if (!(*stash->RA & 1))
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_beq(context *current, StashData *stash)
{
  if (*stash->RA == 0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_blt(context *current, StashData *stash)
{
  if ((signed long) *stash->RA < 0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_ble(context *current, StashData *stash)
{
  if ((signed long) *stash->RA <= 0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_blbs(context *current, StashData *stash)
{
  if (*stash->RA & 1)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_bne(context *current, StashData *stash)
{
  if (*stash->RA)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_bge(context *current, StashData *stash)
{
  if ((signed long) *stash->RA >= 0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_bgt(context *current, StashData *stash)
{
  if ((signed long) *stash->RA > 0)
    stash->taken = 1;
  else 
    stash->taken = 0;
}

emulate_call_pal_callsys(context *current, StashData *stash)
{
	static int tmp;

	if (!(current->wrong_path || current->follow_sync)) {
	  if((tmp = dosyscall(current)) != 0)
	    {
	      /* This thread has exited!! */
	      if(tmp >= 256) {
		current->running = 0;
		printf("**** Exiting at %ld cycles *****\n",
		       cyc);
		schedstats();
		predict_stats();
		print_cstats();
		fflush(0);
		exit(0);
	      }
	    }
	}
}

emulate_call_pal_rduniq(context *current, StashData *stash)
{
  if (!current->wrong_path && !current->follow_sync) {
      current->R[0] = current->hwpcb;
      current->R[0] = dorduniq(current);
    }
}

emulate_call_pal_wruniq(context *current, StashData *stash)
{
  if (!current->wrong_path && !current->follow_sync) {
    current->hwpcb = current->R[REG_A0];
    dowruniq(current);
  }
}

/* I do my best not to decode an instruction (by decode, I
   mean collect all of the static information I possibly
   can about an instruction, including pointers to the
   operands, pointer to the emulation routine, and much
   more) more than once.  Instead, I put everything in
   these stash structures and try and keep track of them.
 */

decode_inst(context *current, StashData *stash)
{
  unsigned int inst;
  unsigned int ra, rb, rc;
  
  inst = rnr_mem_word(current, current->pc); 
  stash->inst = inst;
  ra = (inst >> 21) & 0x1f;
  rb = (inst >> 16) & 0x1f;
  rc = inst & 0x1f;
  stash->RA = &current->R[ra];
  stash->RB = &current->R[rb];
  stash->RC = &current->R[rc];
  stash->offset = inst & 0xffff;
  stash->whichfu = INTEGER;
  stash->branch = 0;
  stash->condbranch = 0;
  stash->syncop = 0;
  stash->srchaz1 = stash->srchaz2 = stash->desthaz = 31;
  current->lasthazard[31] = -1;
  stash->delay = 1;
  stash->smt_primitive = 0;
  
  /*
   * we calculate these two here and discard if unnecessary
   */
  switch(stash->inst >> 26) {
  case LDA:
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->emulate = emulate_lda;
    break;
  case LDAH:
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->emulate = emulate_ldah;
    break;
  case LDQ_U:
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->delay = 2;
    stash->emulate = emulate_ldq_u;
    break;	  
  case STQ_U:
    stash->srchaz1 = ra;
    stash->srchaz2 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_stq_u;
    break;	  
  case INTA:
    if (stash->inst & 0x1000) {
      stash->rbv = (stash->inst >> 13) & 0xff;
      stash->RB = &stash->rbv;
    }
    else {
      stash->srchaz2 = rb;
    }
    stash->desthaz = rc;
    stash->srchaz1 = ra;

    switch (stash->offset >> 5 & 0x7f) {
    case ADDL:
    case ADDLV:
      stash->emulate = emulate_inta_addl;
      break;
    case ADDQ:
    case ADDQV:
      stash->emulate = emulate_inta_addq;
      break;
    case CMPLE:
      stash->delay = 0;
      stash->emulate = emulate_inta_cmple;
      break;
    case CMPULT:
      stash->delay = 0;
      stash->emulate = emulate_inta_cmpult;
      break;
    case S4ADDL:
      stash->emulate = emulate_inta_s4addl;
      break;
    case S4SUBQ:
      stash->emulate = emulate_inta_s4subq;
      break;
    case S8SUBL:
      stash->emulate = emulate_inta_s8subl;
      break;
    case CMPBGE:
      stash->delay = 0;
      stash->emulate = emulate_inta_cmpbge;
      break;
    case CMPLT:
      stash->delay = 0;
      stash->emulate = emulate_inta_cmplt;
      break;
    case SUBL:
    case SUBLV:
      stash->emulate = emulate_inta_subl;
      break;
    case SUBQ:
    case SUBQV:
      stash->emulate = emulate_inta_subq;
      break;
    case S4ADDQ:
      stash->emulate = emulate_inta_s4addq;
      break;
    case S8ADDL:
      stash->emulate = emulate_inta_s8addl;
      break;
    case S8SUBQ:
      stash->emulate = emulate_inta_s8subq;
      break;
    case CMPEQ:
      stash->delay = 0;
      stash->emulate = emulate_inta_cmpeq;
      break;
    case CMPULE:
      stash->delay = 0;
      stash->emulate = emulate_inta_cmpule;
      break;
    case S4SUBL:
      stash->emulate = emulate_inta_s4subl;
      break;
    case S8ADDQ:
      stash->emulate = emulate_inta_s8addq;
      break;
    default:
      emulate_invalid_instruction(current, stash);
      break;
    }
    break;
  case INTL:
    if (stash->inst & 0x1000) {
      stash->rbv = (stash->inst >> 13) & 0xff;
      stash->RB = &stash->rbv;
    }
    else {
      stash->srchaz2 = rb;
    }
    stash->desthaz = rc;
    stash->srchaz1 = ra;

    switch (stash->offset >>5 & 0x7f) {
    case AND:
      stash->emulate = emulate_intl_and;
      break;
    case BIC:
      stash->emulate = emulate_intl_bic;
      break;
    case BIS:
      switch (inst) {
      case SMT_FORK:
	stash->smt_primitive = 1;
        stash->emulate = emulate_intl_bis_smt_fork;
	break;
      case SMT_GETID:
	stash->smt_primitive = 1;
        stash->emulate = emulate_intl_bis_smt_getid;
	break;
      case SMT_CREATESTACK:
	stash->smt_primitive = 1;
        stash->emulate = emulate_intl_bis_smt_createstack;
	break;
      case SMT_HW_LOCK:
	stash->smt_primitive = 1;
	stash->syncop = SMT_HW_LOCK;
	stash->delay = 1;
	stash->whichfu = SYNCH;
#ifdef MFLOW
	stash->srchaz1 = 2;
#else
	stash->srchaz1 = REG_A0;
#endif
        stash->emulate = emulate_intl_bis_smt_hw_lock;
	break;
      case SMT_RELEASE:
	stash->smt_primitive = 1;
	stash->whichfu = SYNCH;
	stash->syncop = SMT_RELEASE;
	stash->delay = 1;
#ifdef MFLOW
	stash->srchaz1 = 2;
#else
	stash->srchaz1 = REG_A0;
#endif
        stash->emulate = emulate_intl_bis_smt_release;
	break;
      case SMT_TERMINATE:
	stash->smt_primitive = 1;
	stash->whichfu = SYNCH;
	stash->syncop = SMT_TERMINATE;
	stash->delay = 4;
        stash->emulate = emulate_intl_bis_smt_terminate;
	break;
      case SMT_START_SIM:
	emulate_intl_bis_smt_start_sim(current, stash);
	break;
      case SMT_PRINT_SIM:
	stash->smt_primitive = 1;
        stash->emulate = emulate_intl_bis_smt_print_sim;
	break;
      case SMT_END_SIM:
	emulate_intl_bis_smt_end_sim(current, stash);
	break;
      case SMT_TAS:
	/* obsolete */
	stash->smt_primitive = 1;
	stash->syncop = SMT_TAS;
	stash->delay = 4;
	stash->whichfu = SYNCH;
        stash->emulate = emulate_intl_bis_smt_tas;
	break;
      default:
        stash->emulate = emulate_intl_bis;
	break;
      }
      break;
    case CMOVEQ:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmoveq;
      break;
    case CMOVLBC:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovlbc;
      break;
    case CMOVLBS:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovlbs;
      break;
    case CMOVGE:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovge;
      break;
    case CMOVGT:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovgt;
      break;
    case CMOVLE:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovle;
      break;
    case CMOVLT:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovlt;
      break;
    case CMOVNE:
      stash->delay = 2;
      stash->emulate = emulate_intl_cmovne;
      break;
    case EQV:
      stash->emulate = emulate_intl_eqv;
      break;
    case ORNOT:
      stash->emulate = emulate_intl_ornot;
      break;
    case XOR:
      stash->emulate = emulate_intl_xor;
      break;
    default:
      emulate_invalid_instruction(current, stash);
      break;
    }
    break;
  case INTS:
    if (stash->inst & 0x1000) {
      stash->rbv = (stash->inst >> 13) & 0xff;
      stash->RB = &stash->rbv;
    }
    else {
      stash->srchaz2 = rb;
    }
    stash->desthaz = rc;
    stash->srchaz1 = ra;

    switch (stash->offset >>5 & 0x7f) {
    case EXTBL:
      stash->emulate = emulate_ints_extbl;
      break;
    case EXTLH:
      stash->emulate = emulate_ints_extlh;
      break;
    case EXTLL:
      stash->emulate = emulate_ints_extll;
      break;
    case EXTQH:
      stash->emulate = emulate_ints_extqh;
      break;
    case EXTQL:
      stash->emulate = emulate_ints_extql;
      break;
    case EXTWH:
      stash->emulate = emulate_ints_extwh;
      break;
    case EXTWL:
      stash->emulate = emulate_ints_extwl;
      break;
    case INSBL:
      stash->emulate = emulate_ints_insbl;
      break;
    case INSLH:
      stash->emulate = emulate_ints_inslh;
      break;
    case INSLL:
      stash->emulate = emulate_ints_insll;
      break;
    case INSQH:
      stash->emulate = emulate_ints_insqh;
      break;
    case INSQL:
      stash->emulate = emulate_ints_insql;
      break;
    case INSWH:
      stash->emulate = emulate_ints_inswh;
      break;
    case INSWL:
      stash->emulate = emulate_ints_inswl;
      break;
    case MSKBL:
      stash->emulate = emulate_ints_mskbl;
      break;
    case MSKLH:
      stash->emulate = emulate_ints_msklh;
      break;
    case MSKLL:
      stash->emulate = emulate_ints_mskll;
      break;
    case MSKQH:
      stash->emulate = emulate_ints_mskqh;
      break;
    case MSKQL:
      stash->emulate = emulate_ints_mskql;
      break;
    case MSKWH:
      stash->emulate = emulate_ints_mskwh;
      break;
    case MSKWL:
      stash->emulate = emulate_ints_mskwl;
      break;
    case SLL:
      stash->emulate = emulate_ints_sll;
      break;
    case SRA:
      stash->emulate = emulate_ints_sra;
      break;	    
    case SRL:
      stash->emulate = emulate_ints_srl;
      break;
    case ZAP:
      stash->emulate = emulate_ints_zap;
      break;
    case ZAPNOT:
      stash->emulate = emulate_ints_zapnot;
      break;
    default:
      emulate_invalid_instruction(current, stash);
      break;
    }
    break;
  case INTM:
    if (stash->inst & 0x1000) {
      stash->rbv = (stash->inst >> 13) & 0xff;
      stash->RB = &stash->rbv;
    }
    else {
      stash->srchaz2 = rb;
    }
    stash->desthaz = rc;
    stash->srchaz1 = ra;

    switch (stash->offset >>5 & 0x7f) {
    case MULL:
    case MULLV:
      stash->delay = 8;
      stash->emulate = emulate_intm_mull;
      break;
    case MULQ:
    case MULQV:
      stash->delay = 16;
      stash->emulate = emulate_intm_mulq;
      break;
    case UMULH:
      stash->delay = 16;
      stash->emulate = emulate_intm_umulh;
      break;
    default:
      emulate_unimplemented_instruction(current, stash);
      break;
    }
    break;
  case FLTL:
    stash->desthaz = rc + FP_OFFSET;
    stash->srchaz1 = rb + FP_OFFSET;
    stash->whichfu = FP;
    stash->delay = 4;

    switch (stash->offset >>5) {
    case CPYS:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_cpys;
      break;
    case CPYSE:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_cpyse;
      break;
    case CPYSN:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_cpysn;
      break;
    case CVTLQ:
      stash->emulate = emulate_fltl_cvtlq;
      break;
    case CVTQL:
    case CVTQLV:
    case CVTQLSV:
      stash->emulate = emulate_fltl_cvtql;
      break;
    case FCMOVEQ:
      /* if Rb = f31, this is a special synch operation, READ_IF_FULL */
#ifdef 0
      if (rb == 31) {
	stash->syncop = READ_FULL;
	stash->whichfu = SYNCH;
	stash->delay = 1;
	stash->emulate = emulate_read_if_full;
      }
      else {
	stash->srchaz2 = ra + FP_OFFSET;
	stash->emulate = emulate_fltl_fcmoveq;
      }
      break;
#else
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_fcmoveq;
      break;
#endif
    case FCMOVGE:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_fcmovge;
      break;
    case FCMOVGT:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_fcmovgt;
      break;
    case FCMOVLE:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_fcmovle;
      break;
    case FCMOVLT:
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_fcmovlt;
      break;
    case FCMOVNE:
#ifdef 0
      /* if Rb = f31, this is a special synch operation, WRITE_AND_FILL */
      if (ra == 31) {
	stash->syncop = WRITE_FILL;
	stash->whichfu = SYNCH;
	stash->srchaz2 = rc;
	stash->desthaz = 63;
	stash->delay = 1;
	stash->emulate = emulate_write_and_fill;
      }
      else {
	stash->srchaz2 = ra + FP_OFFSET;
	stash->emulate = emulate_fltl_fcmovne;
      }
#else
      stash->srchaz2 = ra + FP_OFFSET;
      stash->emulate = emulate_fltl_fcmovne;
#endif
      break;
    case MT_FPCR:
      stash->desthaz = 63;
      stash->emulate = emulate_fltl_mt_fpcr;
      break;
    case MF_FPCR:
      stash->desthaz = ra;
      stash->srchaz1 = 63;
      stash->emulate = emulate_fltl_mf_fpcr;
      break;
    default:
      emulate_invalid_instruction(current, stash);
      break;
    }
    break;
  case FLTI:
    stash->desthaz = rc + FP_OFFSET;
    stash->srchaz1 = rb + FP_OFFSET;
    stash->srchaz2 = ra + FP_OFFSET;
    stash->whichfu = FP;
    stash->delay = 4;

    switch ((stash->offset >>5) & 0x3f) {
    case ADDS:
    case ADDT:
      stash->emulate = emulate_flti_add;
      break;
    case CMPTEQ:
      stash->emulate = emulate_flti_cmpteq;
      break;
    case CMPTLT:
      stash->emulate = emulate_flti_cmptlt;
      break;
    case CMPTLE:
      stash->emulate = emulate_flti_cmptle;
      break;
    case CMPTUN:
      emulate_unimplemented_instruction(current, stash);
      break;
    case CVTQS:
    case CVTQT:
      stash->emulate = emulate_flti_cvtq;
      break;
    case CVTTS:
      stash->emulate = emulate_flti_cvtts;
      break;
    case DIVS: 
      stash->emulate = emulate_flti_divs;
      break;
    case DIVT:
      stash->emulate = emulate_flti_divt;
      break;
    case MULS:
    case MULT:
      stash->emulate = emulate_flti_mul;
      break;
    case SUBS:
    case SUBT:
      stash->emulate = emulate_flti_sub;
      break;
    case CVTTQ:
      stash->emulate = emulate_flti_cvttq;
      break;
    default:
      emulate_unimplemented_instruction(current, stash);
      break;
    }
    break;
  case FLTV:
    emulate_unimplemented_instruction(current, stash);
    break;
  case MISC:
    switch (stash->offset) {
    case TRAPB:
      stash->emulate = emulate_nop;
      break;
    case MB:
      /*
      ** For uniprocessor simulators, the memory barrier instruction
      ** is of no use, and therefore, we don't need to simulate it.
      */
      stash->emulate = emulate_nop;
      break;
    case FETCH:
    case RC : 
    case RS:
    case FETCH_M:
      emulate_unimplemented_instruction(current, stash);
      break;
    case RPCC:
      stash->whichfu = INTLDST;
      emulate_unimplemented_instruction(current, stash);
      break;
    default:
      emulate_invalid_instruction(current, stash);
      break;
    }
    break;
  case JMPJSR:
    stash->delay = 2;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->whichfu = INTEGER;
    stash->branch = 1;

    switch(stash->offset>>14) {
    case JMP:
      stash->emulate = emulate_jmp;
      break;
    case JSR:
      stash->emulate = emulate_jsr;
      break;
    case RETN:
      stash->emulate = emulate_retn;
      break;
    case JSR_COROUTINE:
      stash->emulate = emulate_jsr_coroutine;
      break;
    }
    break;
  case LDF:
    emulate_unimplemented_instruction(current, stash);
    break;
  case LDS:
    stash->delay = 2;
    stash->desthaz = ra + FP_OFFSET;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_lds;
    break;  
  case LDG:
    emulate_unimplemented_instruction(current, stash);
    break;
  case LDT:
    stash->delay = 2;
    stash->desthaz = ra + FP_OFFSET;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_ldt;
    break;	  
  case STF:
    emulate_unimplemented_instruction(current, stash);
    break;
  case STS:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->srchaz2 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_sts;
    break;
  case STG:
    emulate_unimplemented_instruction(current, stash);
    break;
  case STT:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->srchaz2 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_stt;
    break;
  case LDL:
    stash->delay = 2;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_ldl;
    break;	  
  case LDQ:
    stash->delay = 2;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_ldq;
    break;
  case LDL_L:
    stash->smt_primitive = 1;
    stash->syncop = LDL_L;
    stash->delay = 2;
    stash->whichfu = SYNCH;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->emulate = emulate_ldlq_l;
    break;	  
  case LDQ_L:
    stash->smt_primitive = 1;
    stash->syncop = LDQ_L;
    stash->delay = 2;
    stash->whichfu = SYNCH;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->emulate = emulate_ldlq_l;
    break;
  case STL:
    stash->srchaz2 = ra;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_stl;
    break;	  
  case STQ:
    stash->srchaz2 = ra;
    stash->srchaz1 = rb;
    stash->whichfu = INTLDST;
    stash->emulate = emulate_stq;
    break;	  
  case STL_C:	
    stash->smt_primitive = 1;
    stash->syncop = STL_C;
    stash->delay = 1;
    stash->whichfu = SYNCH;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->srchaz2 = ra;
    stash->emulate = emulate_stlq_c;
    break;	  
  case STQ_C:
    stash->smt_primitive = 1;
    stash->syncop = STQ_C;
    stash->delay = 1;
    stash->whichfu = SYNCH;
    stash->desthaz = ra;
    stash->srchaz1 = rb;
    stash->srchaz2 = ra;
    stash->emulate = emulate_stlq_c;
    break;	  
  case BSR: 
    stash->delay = 2;
    stash->desthaz = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->emulate = emulate_bsr;
    break;
  case BR:
    stash->delay = 2;
    stash->desthaz = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->emulate = emulate_br;
    break;
  case FBEQ:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->whichfu = FP;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_fbeq;
    break;
  case FBLT:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->whichfu = FP;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_fblt;
    break;
  case FBLE:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->whichfu = FP;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_fble;
    break;
  case FBNE:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->whichfu = FP;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_fbne;
    break;
  case FBGE:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->whichfu = FP;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_fbge;
    break;
  case FBGT:
    stash->srchaz1 = ra + FP_OFFSET;
    stash->whichfu = FP;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_fbgt;
    break;
  case BLBC:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_blbc;
    break;
  case BEQ:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_beq;
    break;
  case BLT:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_blt;
    break;
  case BLE:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_ble;
    break;
  case BLBS:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_blbs;
    break;
  case BNE:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_bne;
    break;
  case BGE:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_bge;
    break;
  case BGT:
    stash->srchaz1 = ra;
    stash->whichfu = INTEGER;stash->branch = 1;
    stash->condbranch = 1;
    stash->br_target = current->pc + (SEXT(stash->offset) << 2) + 4;
    stash->emulate = emulate_bgt;
    break;
  case CALL_PAL:
    switch(stash->inst & 0x3ffffff) {
    case callsys:
    default:
      stash->whichfu = INTEGER;stash->branch = 1;
      stash->emulate = emulate_call_pal_callsys;
      break;
    case rduniq:
      stash->emulate = emulate_call_pal_rduniq;
      break;
    case wruniq:
      stash->emulate = emulate_call_pal_wruniq;
      break;
    }
    break;
  default:
    if (!(current->wrong_path || current->follow_sync))
      emulate_invalid_instruction(current, stash);
    stash->whichfu = INTEGER;
    stash->desthaz = 30;
    stash->srchaz1 = stash->srchaz2 = 31;
    break;
  }
}

