/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "mem.h"
#include "stack.h"

Segment segment[9];

extern context *thecontexts[];
extern int numcontexts;

#include "memaddr.c"

/*
 * this is to allow passing NULL pointers in system calls without causing
 * illegal address traps.
 */

extern mem_addr mem_address_nullok(context *current, mem_addr addr) {
    Segment s;
    unsigned long diff;

    if(addr == 0) return 0;
redo:
    /*
     * we do the range comparison by an unsigned subtraction
     */
    s = current->segment[(addr >> 28)];
    diff = addr - s.base;
    if(diff <= s.size && s.size > 0)
	return (mem_addr) ((char *)s.segment + diff);
    /* 
     * expand stack if necessary -- segment[8] is always a stack segment
     * if the pointer is within a meg of sp, increase sp.
     */
    if((current->segment[1].base - addr)  < (1 * M)) {
	expand_stack(current);
	goto redo;
    }
    /* we have to set Cause and BadVAddr */
    /*
     * Let's seg fault right here
     */
    *(char *)0 = '\0';
}

/*
 * these have become macros in mem.h 
 * they are retained here only to assist in debugging if necessary
 */

#ifdef SLOW
extern mem_quad read_mem_quad(context *current, mem_addr addr) {
    return *(mem_quad *)mem_address(current, addr);
}
extern mem_word read_mem_word(context *current, mem_addr addr) {
    return *(mem_word *)mem_address(current, addr);
}
extern mem_half read_mem_half(context *current, mem_addr addr) {
    return *(mem_half *)mem_address(current, addr);
}
extern mem_byte read_mem_byte(context *current, mem_addr addr) {
    return *(mem_byte *)mem_address(current, addr);
}
extern void set_mem_quad(context *current, mem_addr addr, mem_quad value) {
    *(mem_quad *)mem_address(current, addr) = value;
}
extern void set_mem_word(context *current, mem_addr addr, mem_word value) {
    *(mem_word *)mem_address(current, addr) = value;
}
extern void set_mem_half(context *current, mem_addr addr, mem_half value) {
    *(mem_half *)mem_address(current, addr) = value;
}
extern void set_mem_byte(context *current, mem_addr addr, mem_byte value) {
    *(mem_byte *)mem_address(current, addr) = value;
}
#endif

/* routines to read simulated application memory, given the simulated
    application's virtual address.  These are safe versions of the
    above, which do alignment and range checking, etc.
 */

extern mem_quad safe_read_mem_quad(context *current, mem_addr addr) {
    Segment s;
    unsigned long diff;

    if ((addr>>28) > 64) 
      return 0;
    s = current->segment[(addr >> 28)];
    diff = (addr - s.base) & ~7;
    if(diff <= s.size && s.size > 0)
	return *(mem_quad *)((char *)s.segment + diff);
    return 0;
}

extern mem_word safe_read_mem_word(context *current, mem_addr addr) {
    Segment s;
    unsigned long diff;

    if ((addr>>28) > 64) 
      return 0;
    s = current->segment[(addr >> 28)];
    diff = (addr - s.base) & ~3;
    if(diff <= s.size && s.size > 0)
	return *(mem_word *)((char *)s.segment + diff);
    return 0;
}

/*  A bunch of various memory management routines.  Address translation
    is done based on a per-context segment map.
*/

static void setsegtype(context *c, mem_addr base, mem_addr *segman, 
		       unsigned long size, MemId id) {
    int seg;

    for(seg = base >> 28; seg <= ((base + size) >> 28); ++seg) {
	c->segment[seg].segmentid = id;
	c->segment[seg].base = base; 
	c->segment[seg].segment = segman;
	c->segment[seg].size = size;
    }
}

unsigned long text_start_address, data_start_address;

static void updatesegmentmap(context *c) {
    setsegtype(c, c->stackbase, (mem_addr *)c->stackseg, c->stacksize, Stack);
    setsegtype(c, text_start_address, (mem_addr *)c->textseg, c->textsize, Text);
    setsegtype(c, data_start_address, (mem_addr *)c->dataseg, c->datasize, Data);
}

makezeroseg() {
  /*setsegtype(current, 0, (mem_addr *) calloc(1024,1), 1024, Data);*/
  }

mem_word 
*make_aligned_segment(unsigned long size, unsigned long offset) {
  unsigned long data;
  mem_word *new_start;
  int shiftamt;

  /* We need to ensure that each segment can begin on a real page
     boundary */
  data = (unsigned long) galloc(size+PAGESIZE);
  if ((data % PAGESIZE) > (offset % PAGESIZE))
    data += PAGESIZE;
  data &= PAGEMASK;
  return (mem_word *) data;
}

extern void make_memory(context *current,
			unsigned long textsz, unsigned long datasz,
			unsigned long textstart, unsigned long datastart,
			int numthreads) {
#ifdef ALIGN_MEM
    current->textseg = make_aligned_segment(textsz,textstart%PAGESIZE);
    current->textsize = textsz;
    current->dataseg = make_aligned_segment(datasz,datastart%PAGESIZE);
    current->datasize = datasz;
    current->stackseg = make_aligned_segment(STACKSZ,0);
    current->stacksize = STACKSZ;
#else
    current->textseg = (mem_word *) galloc(textsz);
    current->textsize = textsz;
    current->dataseg = (mem_word *) galloc(datasz);
    current->datasize = datasz;
    current->stackseg = (mem_word *) galloc(STACKSZ*numthreads);
    current->stacksize = STACKSZ;
#endif
    text_start_address = textstart;
    data_start_address = datastart;
    current->stackbase = stack_allocate_contiguous(current,STACKSZ);
    current->stacktop = current->stackbase+STACKSZ;
    updatesegmentmap(current);
}

extern void share_memory(context *current, context *previous,
			 unsigned long textsz, 
			 unsigned long textstart, unsigned long datastart,
			 int numthreads) {
    current->textseg = previous->textseg;
    current->textsize = previous->textsize;
    current->dataseg = previous->dataseg;
    current->datasize = previous->datasize;
#ifdef OLD
#ifdef ALIGN_MEM
    current->stackseg = make_aligned_segment(STACKSZ,0);
#else
    current->stackseg = (mem_word *) galloc(STACKSZ);
#endif
#else
    current->stackseg = (long) previous->stackseg + numthreads*STACKSZ;
#endif
    current->stacksize = STACKSZ;
    text_start_address = textstart;
    data_start_address = datastart;
    current->stackbase = stack_allocate_contiguous(current, STACKSZ);
    current->stacktop = current->stackbase+STACKSZ-1;
    updatesegmentmap(current);
}

extern void free_memory(context *current) {
    free(current->textseg);
    free(current->dataseg);
    free(current->stackseg);
    stack_free(current->stackbase);
}

/* Expand the data segment by delta */
extern void expand_data(context *current, long delta) {
    int i;
    unsigned long oldsize = current->datasize;

    DEBUGPRINTF("delta %d\n", delta);
    current->datasize += delta;
    current->dataseg = (mem_word *) grealloc(current->dataseg, current->datasize);
    if(delta > 0)
	bzero((char *)current->dataseg + oldsize, delta);
    updatesegmentmap(current);

    /* Update the data segments for contexts which share this data segment */
/*    for (i=0;i<MAXCONTEXTS;i++) {*/
    for (i=0;i<numcontexts;i++) {
      if (thecontexts[i]->masterid == current->masterid) {
	thecontexts[i]->datasize = current->datasize;
	thecontexts[i]->dataseg = current->dataseg;
	updatesegmentmap(thecontexts[i]);
      }
    }
}

/* Expand the stack segment */
void expand_stack(context *current) {
    int oldsize = current->stacksize;

    printf("Expanding stack segment\n");
    current->stacksize += STACKDELTA;
    current->stackseg = (mem_word *) grealloc(current->stackseg, current->stacksize);
    memmove((char*)current->stackseg+STACKDELTA, current->stackseg, oldsize);
    updatesegmentmap(current);
}

void print_memstats(void) {
    int i;
    int j;

    for (j=0;j<numcontexts;j++) {
      if (thecontexts[j]) {
	printf("Context %d (0x%lx):\n", j, thecontexts[j]);
	for (i=0;i<64;i++) {
	  if (thecontexts[j]->segment[i].segment) {
	    printf("   Segment %d (0x%lx) is ", 
		   i, thecontexts[j]->segment[i].segment);
	    switch (thecontexts[j]->segment[i].segmentid) {
	    case Stack:
	      printf("Stack");
	      break;
	    case Text:
	      printf("Text ");
	      break;
	    case Data:
	      printf("Data ");
	      break;
	    }
	    printf(", base = 0x%lx, size = 0x%lx\n", 
		   thecontexts[j]->segment[i].base, 
		   thecontexts[j]->segment[i].size);
	  }
	}
	printf("-------------------\n");
      }
    }
}

static mem_addr copystr(context *current, char *s) {
    mem_addr str_start;
    int i = strlen(s);

    while(i >= 0) {
	snr_mem_byte(current, current->R[30], s[i]);
	current->R[30] = current->R[30] - 1;
	--i;
    }
    str_start = (mem_addr) ((char *) current->R[30] + 1);
    return(str_start);
}

extern unsigned long text_start_address;

extern void setupstack(context *current,
		       int start, int argc, char *argv[], char *env[]) {
    char **p;
    int i = 0, j = 0, k = 0, envj;
    mem_addr addrs[10000];
    mem_addr args[10000];

    /* Put strings on stack: */
/*    current->R[30] = text_start_address - 0x69;*/
    current->R[30] = current->stacktop - 0x69;

    /* Puts environment on the stack.  This actually makes
       results somewhat non-repeatable.  I deal with this by
       essentially making the environment exactly the same
       for every run.  It would probably be better to fix
       it here, by making sure the environment always 
       took a deterministic amount of space.
       */

    for(p = env; *p != NULL; p++)  ;
    for(--p; p >= env; --p) {
      if (strncmp(*p,"TERM=",5) != 0)
	addrs[j++] = copystr(current,*p);
    }
    addrs[j++] = copystr(current,"TERM=vt220");
    addrs[j++] = copystr(current,"LINES=24");
    addrs[j++] = copystr(current,"COLUMNS=80");

    envj = j;
    for (i = start; i < argc; ++i)
	args[k++] = copystr(current,argv[i]);

    /* Round down to double-word boundary */
    current->R[30] = (current->R[30] - 15) & ~15;
  
    snr_mem_quad(current, current->R[30], 0); current->R[30] -= 8;
    /* Build vectors on stack: */
    for (i = 0; i < envj; ++i, current->R[30] -= 8) {
 	/*printf("env %d points to %s\n", i, mem_address(addrs[i])); */
	snr_mem_quad(current, current->R[30], addrs[i]);
    }
#ifdef MFLOW
    current->R[4] = current->R[30] + 8;
#else
    current->R[REG_A2] = current->R[30] + 8;
#endif
    snr_mem_word(current, current->R[30], 0); current->R[30] -= 8;
    for (i = k - 1; i >= 0; i--, current->R[30] -= 8) {
 	/*printf("argv %d %s\n", i, mem_address(args[i])); */
	snr_mem_quad(current, current->R[30], args[i]);
    }
#ifdef MFLOW
    current->R[3]  = current->R[30] + 8;
#else
    current->R[REG_A1]  = current->R[30] + 8;
#endif
    snr_mem_quad(current, current->R[30], argc - start); /* sp points to argc */
#ifdef MFLOW
    current->R[2] = argc - start;
#else
    current->R[REG_A0] = argc - start;
#endif
/*     printf("R[30] = 0x%x %d\n", current->R[30], current->R[30] % 8); */
    /* Need this so that when a new thread is created using an idle
       context, it can reset its stack pointer correctly. */
    current->stacktop = current->R[30];
}
